#!/usr/bin/env python3
#
#  unknown.py
"""
Comparison between project(s) and unknowns.
"""
#
#  Copyright © 2024 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# 3rd party
import pandas  # type: ignore[import-untyped]
from pyms.DPA.Alignment import Alignment

# this package
from libgunshotmatch.project import Project
from libgunshotmatch.utils import create_alignment

# this package
from ._utils import _get_padded_peak_list, _PaddedPeakList

__all__ = ("filter_alignment_to_consolidate", "get_padded_peak_list")


def filter_alignment_to_consolidate(unknown: Project) -> Alignment:
	"""
	Filter peaks from an unknown's ``alignment`` to only those which survived the ``consolidate`` process.

	:param project:
	"""

	assert unknown.consolidated_peaks is not None

	cp_rts = [cp.rt for cp in unknown.consolidated_peaks]

	aligned_peaks = [peak for peak in unknown.alignment.peakpos[0] if peak.rt in cp_rts]

	# Sanity check
	for c_expr_peaks, expr_peaks in zip([aligned_peaks], unknown.alignment.peakpos):
		for peak in c_expr_peaks:
			assert peak in expr_peaks

	# Create new Alignment object
	return create_alignment([aligned_peaks], [unknown.name])


def get_padded_peak_list(unknown: Project, alignment_rts: pandas.DataFrame) -> _PaddedPeakList:
	"""
	Returns a list of consolidated peaks for the unknown sample, based on the between-project alignment.

	:param unknown:
	:param alignment_rts: Pandas DataFrame giving retention times for the peak alignment.
		The output of :meth:`~pyms.DPA.Alignment.Alignment.get_peak_alignment`.
	"""

	rts = list(alignment_rts[unknown.name])

	return _get_padded_peak_list(unknown, rts)



#!/usr/bin/env python3
#
#  __init__.py
"""
Methods for GunShotMatch analysis.
"""
#
#  Copyright © 2020-2023 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

try:
	# stdlib
	import tomllib  # type: ignore[import-not-found]
except ImportError:
	import tomli as tomllib

# stdlib
import json
from typing import Any, Dict, List, Mapping, Optional, Set, Tuple, Type, TypeVar, Union

# 3rd party
import attr
import tomli_w

# this package
from libgunshotmatch.method._fields import (
		Boolean,
		Integer,
		Number,
		String,
		convert_crop_mass_range,
		convert_rt_range,
		convert_sg_window,
		default_base_peak_filter
		)
from libgunshotmatch.utils import _fix_init_annotations, _to_list

__all__ = (
		"MethodBase",
		"Method",
		"IntensityMatrixMethod",
		"PeakDetectionMethod",
		"PeakFilterMethod",
		"AlignmentMethod",
		"ConsolidateMethod",
		"SavitzkyGolayMethod",
		)

_MB = TypeVar("_MB", bound="MethodBase")


class MethodBase:
	"""
	Base class for methods.
	"""

	@classmethod
	def from_dict(cls: Type["MethodBase"], method: Mapping[str, Any]) -> "MethodBase":
		"""
		Construct a :class:`~.MethodBase` from a dictionary.

		:param method:
		"""

		return cls(**method)

	def to_dict(self) -> Dict[str, Any]:
		"""
		Convert a :class:`~.MethodBase` to a dictionary.
		"""

		return attr.asdict(self, recurse=True)  # type: ignore[arg-type]  # This is an ABC but not an attrs class.

	@classmethod
	def _coerce(cls: Type[_MB], method: Any) -> _MB:
		if isinstance(method, cls):
			return method
		elif isinstance(method, Mapping):
			return cls(**method)
		else:

			class_name = cls.__name__
			if class_name.startswith("aeiouAEIOU"):
				raise TypeError(f"Cannot convert {type(method).__name__} to an {class_name}")
			# TODO: edge cases
			else:
				raise TypeError(f"Cannot convert {type(method).__name__} to a {class_name}")


@_fix_init_annotations
@attr.define
class SavitzkyGolayMethod(MethodBase):
	"""
	Method parameters for the Savitzky-Golay filter.

	.. versionadded:: 0.3.0
	"""

	#: Whether to perform Savitzky-Golay smoothing.
	enable: bool = Boolean.field(default=True)

	window: Union[str, int] = attr.field(default=7, converter=convert_sg_window)
	"""
	The window size for the Savitzky-Golay filter.

	Either a number of scans or a must be the form ``'<NUMBER>s'`` or ``'<NUMBER>m'``,
	specifying a time in seconds or minutes, respectively.
	"""

	#: The degree of the fitting polynomial for the Savitzky-Golay filter.
	degree: int = Integer.field(default=2)


def _convert_sg_method(method: Union[bool, "SavitzkyGolayMethod", Dict[str, Any]]) -> "SavitzkyGolayMethod":
	if isinstance(method, bool):
		return SavitzkyGolayMethod()
	elif isinstance(method, SavitzkyGolayMethod):
		return method
	else:
		return SavitzkyGolayMethod(**method)


@_fix_init_annotations
@attr.define
class IntensityMatrixMethod(MethodBase):
	"""
	Method used for constructing an intensity matrix from a datafile.
	"""

	#: The range of masses to which the GC-MS data should be limited to.
	crop_mass_range: Optional[Tuple[int, int]] = attr.field(default=(50, 500), converter=convert_crop_mass_range)

	# Whether to perform Savitzky-Golay smoothing.
	#: Settings for Savitzky-Golay smoothing.
	savitzky_golay: SavitzkyGolayMethod = attr.field(
			default=SavitzkyGolayMethod(),
			converter=_convert_sg_method,
			)

	#: Whether to perform Tophat baseline correction.
	tophat: bool = Boolean.field(default=True)

	#: The structure size for Tophat baseline correction.
	tophat_structure_size: str = String.field(default="1.5m")


@_fix_init_annotations
@attr.define
class PeakDetectionMethod(MethodBase):
	"""
	Method used for Biller-Biemann peak detection.
	"""

	#: Number of scans over which to consider a maxima to be a peak.
	points: int = Integer.field(default=10)

	#: Number of scans to combine in a single peak from to compensate for spectra skewing.
	scans: int = Integer.field(default=1)


@_fix_init_annotations
@attr.define
class PeakFilterMethod(MethodBase):
	"""
	Method used for peak filtering.
	"""

	#: Whether to perform automatic noise filtering of the peak list.
	noise_filter: bool = Boolean.field(default=True)

	#: The minimum number of ions that must have intensities above the noise floor, otherwise the peak is excluded.
	noise_threshold: int = Integer.field(default=2)

	# TODO: non-integer binned data
	#: Peaks whose base peak is at one of the listed masses (m/z) are excluded.
	base_peak_filter: Set[int] = attr.field(
			default=attr.Factory(default_base_peak_filter),
			converter=set,
			validator=attr.validators.instance_of(set)
			)

	#: Optional retention time range (in minutes) to filter the peak list to.
	rt_range: Optional[Tuple[float, float]] = attr.field(default=None, converter=convert_rt_range)


@_fix_init_annotations
@attr.define
class AlignmentMethod(MethodBase):
	"""
	Method used for peak alignment.
	"""

	#: Retention time tolerance parameter for pairwise alignments.
	rt_modulation: float = Number.field(default=2.5)

	#: Gap parameter for pairwise alignments.
	gap_penalty: float = Number.field(default=0.3)

	min_peaks: int = Integer.field(default=1)
	"""
	Minimum number of peaks required for the alignment position to survive filtering.

	If set to ``-1`` the number of repeats in the project are used.
	"""

	#: Number of peaks (starting with the largest) to include in the output.
	top_n_peaks: int = Integer.field(default=80)

	#: Minimum area of peaks to include in the output.
	min_peak_area: float = Number.field(default=0.0)


@_fix_init_annotations
@attr.define
class ConsolidateMethod(MethodBase):
	"""
	Method used for consolidation (finding most likely identity for aligned peaks).

	:param min_appearances: Number of times the hit must appear across the individual aligned peaks.
		Consolidated peaks where the most common hit appears fewer times than this will be excluded.
		If set to ``-1`` the number of instances of the peak in the project are used.

	.. versionchanged:: 0.2.0  Added the ``min_appearances`` argument.
	"""

	name_filter: List[str] = attr.field(converter=_to_list, default=attr.Factory(list))
	"""
	List of glob-style matches for compound names.

	Consolidated peaks matching any of these will be excluded.
	"""

	min_match_factor: int = Integer.field(default=600)
	"""
	Minimum average match factor.

	Consolidated peaks with an average match factor below this will be excluded.
	"""

	min_appearances: int = Integer.field(default=-1)
	"""
	Number of times the hit must appear across the individual aligned peaks.

	Consolidated peaks where the most common hit appears fewer times than this will be excluded.

	If set to ``-1`` the number of instances of the peak in the project are used.

	.. versionadded:: 0.2.0
	"""


# target_range = 4.0,37.0


def _submethod_field(submethod_type: Type[_MB]) -> _MB:
	# Actually returns attr.Attribute, but mypy doesn't like it
	return attr.field(factory=submethod_type, converter=submethod_type._coerce)


@_fix_init_annotations
@attr.define
class Method(MethodBase):
	"""
	Overall GunShotMatch method.

	.. latex:vspace:: 4mm
	"""

	#: Method used for constructing an intensity matrix from a datafile.
	intensity_matrix: IntensityMatrixMethod = _submethod_field(IntensityMatrixMethod)

	#: Method used for Biller-Biemann peak detection.
	peak_detection: PeakDetectionMethod = _submethod_field(PeakDetectionMethod)

	#: Method used for peak filtering.
	peak_filter: PeakFilterMethod = _submethod_field(PeakFilterMethod)

	#: Method used for peak alignment.
	alignment: AlignmentMethod = _submethod_field(AlignmentMethod)

	#: Method used for consolidation (finding most likely identity for aligned peaks).
	consolidate: ConsolidateMethod = _submethod_field(ConsolidateMethod)

	@classmethod
	def from_toml(cls: Type["Method"], toml_string: str) -> "Method":
		"""
		Parse a :class:`~.Method` from a TOML string.

		:param toml_string:
		"""

		parsed_toml = tomllib.loads(toml_string)
		return cls(**parsed_toml["method"])

	@classmethod
	def from_json(cls: Type["Method"], json_string: str) -> "Method":
		"""
		Parse a :class:`~.Method` from a JSON string.

		:param json_string:
		"""

		parsed_json = json.loads(json_string)
		return cls(**parsed_json["method"])

	def to_toml(self) -> str:
		"""
		Convert a :class:`~.Method` to a TOML string.
		"""

		return tomli_w.dumps({"method": self.to_dict()})



#!/usr/bin/env python3
#
#  datafile.py
"""
Represents a parsed datafile.
"""
#
#  Copyright © 2020-2023 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import getpass
import os
import socket
from datetime import datetime
from statistics import mean, median
from typing import Any, Dict, List, Mapping, NamedTuple, Optional, Tuple, Type, Union

# 3rd party
import attr
import pyms.Noise.SavitzkyGolay
import pyms.TopHat
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.typing import PathLike
from enum_tools import IntEnum
from pyms.GCMS.Class import GCMS_data
from pyms.IntensityMatrix import IntensityMatrix, build_intensity_matrix_i

# this package
from libgunshotmatch import gzip_util
from libgunshotmatch.method import SavitzkyGolayMethod
from libgunshotmatch.peak import PeakList, QualifiedPeak, _to_peak_list, peak_from_dict

__all__ = ("Datafile", "FileType", "Repeat", "get_info_from_gcms_data", "GCMSDataInfo")


class FileType(IntEnum):
	"""
	Represents the input datafile types supported by PyMassSpec.
	"""

	#: JCAMP-DX (https://iupac.org/wp-content/uploads/2021/08/JCAMP-DX_MS_1994.pdf)
	JDX = 0

	#: mzML (https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3073315/)
	MZML = 1

	#: Analytical Data Interchange Format for Mass Spectrometry (https://www.astm.org/e1947-98r22.html)
	ANDI = 2

	@classmethod
	def _attrs_convert(cls: Type["FileType"], value: int) -> "FileType":
		if value in iter(cls):  # type: ignore[operator]
			return cls(value)

		raise ValueError(f"Unrecognised file format {value}")


@attr.define
class Datafile:
	"""
	Represents a single datafile in a project.

	:default user: taken from the currently logged-in user
	:default device: taken from the current device's hostname
	:default date_created: is the current date and time
	:default date_modified: is the current date and time
	"""

	#: The name of the :class:`~.Datafile`.
	name: str
	#: The filename of the file the :class:`~.Datafile` was created from.
	original_filename: str
	#: The filetype of the file the :class:`~.Datafile` was created from.
	original_filetype: FileType = attr.field(
			converter=FileType._attrs_convert,  # type: ignore[misc]  # doesn't like the converter
			)
	#: A description of the :class:`~.Datafile`.
	description: str = attr.field(default='')
	#: PyMassSpec :class:`~pyms.IntensityMatrix.IntensityMatrix` object.
	intensity_matrix: Optional[IntensityMatrix] = attr.field(default=None)

	#: The user who created the :class:`~.Datafile`.
	user: str = attr.field(factory=getpass.getuser)
	#: The device that created the :class:`~.Datafile`.
	device: str = attr.field(factory=socket.gethostname)
	#: The date and time the :class:`~.Datafile` was created.
	date_created: datetime = attr.field(factory=datetime.now)
	#: The date and time the :class:`~.Datafile` was last modified.
	date_modified: datetime = attr.field(factory=datetime.now)
	#: File format version
	version: int = attr.field(default=1)

	@classmethod
	def new(cls: Type["Datafile"], name: str, filename: PathLike) -> "Datafile":
		"""
		Construct a new :class:`~.Datafile` from a file.

		:param name: The name of the Datafile
		:param filename:
		"""

		date_modified = date_created = datetime.now()

		filename_p = PathPlus(filename)

		if filename_p.suffix.lower() == ".jdx":
			original_filetype = FileType.JDX
		elif filename_p.suffix.lower() == ".cdf":
			original_filetype = FileType.ANDI
		elif filename_p.suffix.lower() == ".mzml":
			original_filetype = FileType.MZML
		else:
			raise ValueError(f"Unrecognised file format for file {filename}")

		return cls(
				name=name,
				date_created=date_created,
				date_modified=date_modified,
				original_filename=filename_p.resolve().as_posix(),
				original_filetype=original_filetype,
				)

	def load_gcms_data(self, filename: Optional[PathLike] = None) -> GCMS_data:
		"""
		Load GC-MS data from the datafile.

		:param filename: Alternative filename to load the data from. Useful if the file has moved since the :class:`~.Datafile` was created.

		.. versionchanged:: 0.4.0  Added the ``filename`` attribute.
		"""

		filename = PathPlus(filename or self.original_filename)

		if self.original_filetype == FileType.JDX:
			# 3rd party
			from pyms.GCMS.IO.JCAMP import JCAMP_reader
			gcms_data = JCAMP_reader(filename)

		elif self.original_filetype == FileType.MZML:
			# 3rd party
			from pyms.GCMS.IO.MZML import mzML_reader
			gcms_data = mzML_reader(filename)

		elif self.original_filetype == FileType.ANDI:
			# 3rd party
			from pyms.GCMS.IO.ANDI import ANDI_reader
			gcms_data = ANDI_reader(filename)

		else:
			# Shouldn't get here due to filetype validation at attrs' end
			raise ValueError(f"Unrecognised file format {self.original_filetype._name_}")

		return gcms_data

	def prepare_intensity_matrix(
			self,
			gcms_data: GCMS_data,
			savitzky_golay: Union[bool, SavitzkyGolayMethod] = True,
			tophat: bool = True,
			tophat_structure_size: str = "1.5m",  # Ignored if tophat=False
			crop_mass_range: Optional[Tuple[float, float]] = None,
			) -> IntensityMatrix:
		"""
		Build an :class:`~pyms.IntensityMatrix.IntensityMatrix` for the datafile.

		:param gcms_data:
		:param savitzky_golay: Whether to perform Savitzky-Golay smoothing.
		:param tophat: Whether to perform Tophat baseline correction.
		:param tophat_structure_size: The structure size for Tophat baseline correction.
		:param crop_mass_range: The range of masses to which the GC-MS data should be limited to.

		"""

		intensity_matrix = build_intensity_matrix_i(gcms_data)

		# Show the m/z of the maximum and minimum bins
		print(f" Minimum m/z bin: {intensity_matrix.min_mass}")
		print(f" Maximum m/z bin: {intensity_matrix.max_mass}")

		# Crop masses
		if crop_mass_range is not None:
			# None means don't crop

			# min_mass, max_mass = 50, 400
			min_mass, max_mass = crop_mass_range

			# Catch case where numbers are flipped
			if min_mass >= max_mass:
				raise ValueError(
						'\n'.join([
								"Cannot crop mass range when `max mass` is less than `min mass`.\n'",
								"Did you put the numbers the wrong way around? The expected order is (<min>, <max>)",
								])
						)

			if intensity_matrix.min_mass is not None and min_mass < intensity_matrix.min_mass:
				min_mass = intensity_matrix.min_mass
			if intensity_matrix.max_mass is not None and max_mass > intensity_matrix.max_mass:
				max_mass = intensity_matrix.max_mass

			intensity_matrix.crop_mass(min_mass, max_mass)

		# Perform Data filtering
		n_mz = intensity_matrix.size[1]

		# Iterate over each IC in the intensity matrix
		for index in range(n_mz):
			# print("\rWorking on IC#", index+1, '  ',end='')
			ic = intensity_matrix.get_ic_at_index(index)

			if savitzky_golay:
				# Perform Savitzky-Golay smoothing.
				# Note that Turbomass does not use smoothing for qualitative method.
				if isinstance(savitzky_golay, SavitzkyGolayMethod):
					ic = pyms.Noise.SavitzkyGolay.savitzky_golay(
							ic,
							window=savitzky_golay.window,
							degree=savitzky_golay.degree,
							)
				else:
					ic = pyms.Noise.SavitzkyGolay.savitzky_golay(ic)

			if tophat:
				# Perform Tophat baseline correction
				# Top-hat baseline Correction seems to bring down noise,
				#  		retaining shapes, but keeps points on actual peaks
				# ic = tophat(ic, struct=method.tophat_struct)
				ic = pyms.TopHat.tophat(ic, struct=tophat_structure_size)

			# Set the IC in the intensity matrix to the filtered one
			intensity_matrix.set_ic_at_index(index, ic)

		self.intensity_matrix = intensity_matrix
		return intensity_matrix

	def to_dict(self) -> Dict[str, Any]:
		"""
		Returns a dictionary representation of this :class:`~.Datafile`.

		All keys are native, JSON-serializable, Python objects.
		"""

		if self.intensity_matrix is None:
			im_as_dict = None
		else:
			im_as_dict = {
					"times": self.intensity_matrix.time_list,
					"masses": self.intensity_matrix.mass_list,
					"intensities": self.intensity_matrix.intensity_array.tolist(),
					}

		return {
				"name": self.name,
				"original_filename": self.original_filename,
				"original_filetype": int(self.original_filetype),
				"description": self.description,
				"intensity_matrix": im_as_dict,
				"user": self.user,
				"device": self.device,
				"date_created": self.date_created.isoformat(),
				"date_modified": self.date_modified.isoformat(),
				"version": self.version,
				}

	@classmethod
	def from_dict(cls: Type["Datafile"], d: Mapping[str, Any]) -> "Datafile":
		"""
		Construct a :class:`~.Datafile` from a dictionary.

		:param d:
		"""

		im_as_dict = d["intensity_matrix"]
		if im_as_dict is None:
			intensity_matrix = None
		else:
			intensity_matrix = IntensityMatrix(
					time_list=im_as_dict["times"],
					mass_list=im_as_dict["masses"],
					intensity_array=im_as_dict["intensities"],
					)

		optional_keys = {}
		if "user" in d:
			optional_keys["user"] = d["user"]
		if "device" in d:
			optional_keys["device"] = d["device"]
		if "date_created" in d:
			optional_keys["date_created"] = datetime.fromisoformat(d["date_created"])
		if "date_modified" in d:
			optional_keys["date_modified"] = datetime.fromisoformat(d["date_modified"])
		if "version" in d:
			optional_keys["version"] = d["version"]

		return cls(
				name=d["name"],
				original_filename=d["original_filename"],
				original_filetype=FileType(d["original_filetype"]),
				description=d["description"],
				intensity_matrix=intensity_matrix,
				**optional_keys,
				)

	def export(self, output_dir: PathLike) -> str:
		"""
		Export as a ``.gsmd`` file and return the output filename.
		"""

		export_filename = os.path.join(output_dir, f"{self.name}.gsmd")
		gzip_util.write_gzip_json(export_filename, self.to_dict(), indent=None)
		return export_filename

	@classmethod
	def from_file(cls: Type["Datafile"], filename: PathLike) -> "Datafile":
		"""
		Parse a ``gsmd`` file.

		:param filename: The input filename.
		"""

		as_dict: Dict[str, Any] = gzip_util.read_gzip_json(filename)  # type: ignore[assignment]
		return Datafile.from_dict(as_dict)


class GCMSDataInfo(NamedTuple):
	"""
	Represents information about a :class:`~pyms.GCMS.Class.GCMS_data` object returned by :func:`get_info_from_gcms_data`.
	"""

	#: The minimum and maximum retention times.
	rt_range: Tuple[float, float]

	#: The average time step between scans.
	time_step: float

	#: The standard deviation of the time steps between scans.
	time_step_stdev: float

	#: The total number of scans.
	n_scans: int

	#: The minimum and maximum mass (*m/z*) values.
	mz_range: Tuple[float, float]

	#: The mean average number of masses per scan.
	num_mz_mean: float

	#: The median number of masses per scan.
	num_mz_median: float


def get_info_from_gcms_data(gcms_data: GCMS_data) -> GCMSDataInfo:
	"""
	Returns a information about the data in a :class:`pyms.GCMS.Class.GCMS_data` object.
	"""

	# TODO: within pyms make read only properties for these private attributes

	# calculate median number of m/z values measured per scan
	scan_list = gcms_data.scan_list

	n_list = [len(scan) for scan in scan_list]
	n_mz_mean = mean(n_list)
	n_mz_median = median(n_list)

	min_mass = gcms_data.min_mass or -1
	max_mass = gcms_data.max_mass or -1
	info = GCMSDataInfo(
			rt_range=(gcms_data._min_rt / 60, gcms_data._max_rt / 60),
			time_step=gcms_data._time_step,
			time_step_stdev=gcms_data._time_step_std,
			n_scans=len(scan_list),
			mz_range=(min_mass, max_mass),
			num_mz_mean=n_mz_mean,
			num_mz_median=n_mz_median,
			)

	return info


@attr.define
class Repeat:
	"""
	Represents a repeat sample in a project, constructed from a datafile.

	:default user: taken from the currently logged-in user.
	:default device: taken from the current device's hostname.
	:default date_created: is the current date and time.
	:default date_modified: is the current date and time.
	"""

	#: The :class:`~.Datafile` for this repeat.
	datafile: Datafile

	peaks: PeakList = attr.field(converter=_to_peak_list)

	#: Peaks containing identities from library search. This is usually populated after peak alignment.
	qualified_peaks: Optional[List[QualifiedPeak]] = attr.field(default=None)

	#: The user who created the :class:`~.Repeat`.
	user: str = attr.field(factory=getpass.getuser)
	#: The device that created the :class:`~.Repeat`.
	device: str = attr.field(factory=socket.gethostname)
	#: The date and time the :class:`~.Repeat` was created.
	date_created: datetime = attr.field(factory=datetime.now)
	#: The date and time the :class:`~.Repeat` was last modified.
	date_modified: datetime = attr.field(factory=datetime.now)
	#: File format version
	version: int = attr.field(default=1)

	@property
	def name(self) -> str:
		"""
		The name of the :class:`~.Datafile`.

		:rtype:

		.. versionadded:: 0.4.0
		"""

		return self.datafile.name

	def to_dict(self) -> Dict[str, Any]:
		"""
		Returns a dictionary representation of this :class:`~.Repeat`.

		All keys are native, JSON-serializable, Python objects.
		"""

		if self.qualified_peaks is None:
			qualified_peaks_as_list = None
		else:
			qualified_peaks_as_list = [qp.to_dict() for qp in self.qualified_peaks]

		return {
				"datafile": self.datafile.to_dict(),
				"peaks": self.peaks.to_list(),
				"qualified_peaks": qualified_peaks_as_list,
				"user": self.user,
				"device": self.device,
				"date_created": self.date_created.isoformat(),
				"date_modified": self.date_modified.isoformat(),
				"version": self.version,
				}

	@classmethod
	def from_dict(cls: Type["Repeat"], d: Mapping[str, Any]) -> "Repeat":
		"""
		Construct a :class:`~.Repeat` from a dictionary.

		:param d:
		"""

		datafile = Datafile.from_dict(d["datafile"])

		peaks = PeakList(peak_from_dict(peak) for peak in d["peaks"])
		peaks.datafile_name = datafile.name

		qualified_peaks_as_list = d["qualified_peaks"]
		if qualified_peaks_as_list is None:
			qualified_peaks = None
		else:
			qualified_peaks = [QualifiedPeak.from_dict(peak) for peak in d["qualified_peaks"]]

		optional_keys = {}
		if "user" in d:
			optional_keys["user"] = d["user"]
		if "device" in d:
			optional_keys["device"] = d["device"]
		if "date_created" in d:
			optional_keys["date_created"] = datetime.fromisoformat(d["date_created"])
		if "date_modified" in d:
			optional_keys["date_modified"] = datetime.fromisoformat(d["date_modified"])
		if "version" in d:
			optional_keys["version"] = d["version"]

		return cls(
				datafile=datafile,
				peaks=peaks,
				qualified_peaks=qualified_peaks,
				**optional_keys,
				)

	def export(self, output_dir: PathLike) -> str:
		"""
		Export as a ``.gsmr`` file and return the output filename.

		.. versionadded:: 0.4.0
		"""

		export_filename = os.path.join(output_dir, f"{self.name}.gsmr")
		gzip_util.write_gzip_json(export_filename, self.to_dict(), indent=None)
		return export_filename

	@classmethod
	def from_file(cls: Type["Repeat"], filename: PathLike) -> "Repeat":
		"""
		Parse a ``gsmr`` file.

		:param filename: The input filename.

		:rtype:

		.. versionadded:: 0.4.0
		"""

		as_dict: Dict[str, Any] = gzip_util.read_gzip_json(filename)  # type: ignore[assignment]
		return cls.from_dict(as_dict)



#!/usr/bin/env python3
#
#  _fields.py
"""
Internal attrs field helpers.
"""
#
#  Copyright © 2020-2023 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import Dict, Mapping, Optional, Union

# 3rd party
import pandas  # type: ignore[import-untyped]
from pyms_nist_search import ReferenceData


def _attrs_convert_cas(cas: str) -> str:
	if cas == "0-00-0":
		cas = "---"

	return cas


_reference_data_error_msg = ''.join([
		"'reference_data' must be a `pyms_nist_search.ReferenceData` object,",
		"a dictionary representing a `ReferenceData` object,",
		"or `None`",
		])


def _attrs_convert_reference_data(reference_data: Union[Dict, ReferenceData, None]) -> Optional[ReferenceData]:

	if reference_data is None:
		return None

	elif isinstance(reference_data, ReferenceData):
		return reference_data

	elif isinstance(reference_data, dict):
		expected_keys = {
				"name",
				"cas",
				"formula",
				"contributor",
				"nist_no",
				"id",
				"mw",
				"mass_spec",
				"synonyms",
				"exact_mass"
				}
		if set(reference_data.keys()) != expected_keys:
			# print(set(reference_data.keys()))
			raise TypeError(_reference_data_error_msg)
		else:
			return ReferenceData(**reference_data)

	else:
		raise TypeError(_reference_data_error_msg)


def _attrs_convert_ms_comparison(
		ms_comparison: Union[Mapping[str, float], pandas.Series, None],
		) -> pandas.Series:

	if ms_comparison is None:
		return pandas.Series()
	elif isinstance(ms_comparison, pandas.Series):
		return ms_comparison
	elif isinstance(ms_comparison, Mapping):
		return pandas.Series(ms_comparison)
	else:
		raise TypeError("'ms_comparison' must be a mapping or a pandas.Series")



#!/usr/bin/env python3
#
#  search.py
"""
Library search functions.
"""
#
#  Copyright © 2020-2023 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import Iterable, List

# 3rd party
import pandas  # type: ignore[import-untyped]
import pyms_nist_search
from pyms.Peak.Class import Peak

# this package
from libgunshotmatch.peak import QualifiedPeak
from libgunshotmatch.utils import round_rt

__all__ = ("identify_peaks", )


def identify_peaks(
		engine: pyms_nist_search.Engine,
		peaks_to_identify: Iterable[float],
		peak_list: List[Peak],
		n_hits: int = 10,
		verbose: bool = False,
		) -> List[QualifiedPeak]:
	"""
	Identify the peaks in ``peak_list`` where their retention times are in ``peaks_to_identify``.

	:param engine:
	:param peaks_to_identify: List of retention times of peaks to identify.
	:param peak_list:
	:param n_hits: The number of hits to return for each peak.
	:param verbose: Enable debug logging
	"""

	# TODO: Shared engine between multiple calls to identify_peaks
	# (perhaps wrap this function in a class)

	# Convert float retention times to Decimal
	# rt_list = [rounders(rt, "0.0000000000") for rt in target_times]
	target_times = pandas.Series(peaks_to_identify).apply(round_rt)

	# Remove NaN values
	rt_list = [rt for rt in target_times if not rt.is_nan()]

	# Sort smallest to largest
	rt_list.sort()

	# # Obtain area for each peak
	# peak_area_list = get_area_list(self.peak_list)
	peaks = []

	# Filter to those peaks present in all samples, by UID
	for peak in peak_list:

		rounded_rt = round_rt(peak.rt / 60)

		if rounded_rt in rt_list:
			qualified_peak = QualifiedPeak.from_peak(peak)
			qualified_peak.peak_number = target_times[target_times == rounded_rt].index[0]

			ms = qualified_peak.mass_spectrum

			if verbose:
				print(f"Identifying peak at rt {rounded_rt} minutes...")

			hit_list = engine.full_spectrum_search(ms, n_hits)

			# Add search results to peak
			for hit in hit_list:
				qualified_peak.hits.append(hit)

			peaks.append(qualified_peak)

	return peaks



#!/usr/bin/env python3
#
#  __init__.py
"""
Comparison between projects and unknowns.

The two submodules, :mod:`~.comparison.projects` and :mod:`~.comparison.unknowns`,
provide identical APIs with different internals to handle reference projects (containing two or more repeats)
and unknown samples (from a single datafile).

.. versionadded:: 0.8.0
"""
#
#  Copyright © 2024 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import List, Sequence, Tuple, Union

# 3rd party
from pyms.DPA.Alignment import Alignment
from pyms.DPA.PairwiseAlignment import PairwiseAlignment, align_with_tree

# this package
from libgunshotmatch.project import Project

# this package
from . import projects, unknowns
from ._utils import _PaddedPeakList

__all__ = ("align_projects", "get_padded_peak_lists")

# Aliases to prevent clashes with argument names
_projects_mod = projects
_unknowns_mod = unknowns


def align_projects(
		projects: Union[Sequence[Project], Project] = (),
		unknowns: Union[Sequence[Project], Project] = (),
		D: float = 2.5,
		gap: float = 0.3,
		) -> Alignment:
	"""
	Align multiple projects and/or unknowns.

	:param projects:
	:param unknowns:
	:param D: Retention time tolerance for pairwise alignments (in seconds).
	:param gap: Gap penalty for pairwise alignments.

	:rtype:

	.. versionchanged:: 0.9.0

		* Added ``D`` and ``gap`` arguments.
		* ``projects`` and ``unknowns`` can now be a single :class:`~.Project`.
	"""

	if isinstance(projects, Project):
		projects = [projects]

	if isinstance(unknowns, Project):
		unknowns = [unknowns]

	project_alignments = map(_projects_mod.filter_alignment_to_consolidate, projects)
	unknown_alignments = map(_unknowns_mod.filter_alignment_to_consolidate, unknowns)

	pwa = PairwiseAlignment([*project_alignments, *unknown_alignments], D=float(D), gap=float(gap))
	return align_with_tree(pwa)


def get_padded_peak_lists(
		alignment: Alignment,
		projects: Union[Sequence[Project], Project] = (),
		unknowns: Union[Sequence[Project], Project] = (),
		) -> Tuple[List[_PaddedPeakList], List[_PaddedPeakList]]:
	"""
	Pad the consolidated peak lists in each project/unknown, from the given between-project alignment.

	:param alignment:
	:param projects:
	:param unknowns:

	.. versionchanged:: 0.9.0  ``projects`` and ``unknowns`` can now be a single :class:`~.Project`.
	"""

	if isinstance(projects, Project):
		projects = [projects]

	if isinstance(unknowns, Project):
		unknowns = [unknowns]

	data = alignment.get_peak_alignment(require_all_expr=False, minutes=False)

	projects_padded_cp = [_projects_mod.get_padded_peak_list(p, data) for p in projects]
	unknowns_padded_cp = [_unknowns_mod.get_padded_peak_list(p, data) for p in unknowns]
	return projects_padded_cp, unknowns_padded_cp



#!/usr/bin/env python3
#
#  consolidate.py
"""
Functions for combining peak identifications across aligned peaks into a single set of results.
"""
#
#  Copyright © 2020-2023 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from collections import Counter, defaultdict
from fnmatch import fnmatch
from itertools import permutations
from multiprocessing import Pool
from typing import Any, Dict, Iterator, List, Mapping, MutableSequence, Optional, Tuple, Type, Union, cast

# 3rd party
import attr
import numpy
import pandas  # type: ignore[import-untyped]
import pyms_nist_search
from pyms.DPA.Alignment import Alignment
from pyms.Spectrum import MassSpectrum
from pyms_nist_search import ReferenceData, SearchResult

# this package
from libgunshotmatch.consolidate._fields import (
		_attrs_convert_cas,
		_attrs_convert_ms_comparison,
		_attrs_convert_reference_data
		)
from libgunshotmatch.consolidate._spectra import PermsListType, _map_func
from libgunshotmatch.method import ConsolidateMethod
from libgunshotmatch.method._fields import Integer
from libgunshotmatch.peak import QualifiedPeak
from libgunshotmatch.utils import _fix_init_annotations, _to_list

__all__ = (
		"ConsolidatedPeak",
		"ConsolidatedSearchResult",
		"match_counter",
		"pairwise_ms_comparisons",
		"ConsolidatedPeakFilter",
		"InvertedFilter",
		"combine_spectra",
		)


@attr.define(frozen=True)
class ConsolidatedSearchResult:
	"""
	Represents a candidate compound for a peak.

	This is determined from a set of :class:`SearchResults <pyms_nist_search.search_result.SearchResult>` for a set of aligned peaks.
	"""

	# TODO:  Not currently copying hit_prob from SearchResult

	#: The name of the candidate compound.
	name: str

	#: The CAS number of the compound.
	cas: str = attr.field(converter=_attrs_convert_cas)

	mf_list: List[int] = attr.field(default=attr.Factory(list))
	"""
	List of Match Factors comparing the mass spectrum of the peak with the reference spectrum in each aligned peak.

	Will contain NaN where the compound was not in the hit list for a peak.
	"""

	rmf_list: List[int] = attr.field(default=attr.Factory(list))
	"""
	List of Reverse Match Factors comparing the reference spectrum with the spectrum for each aligned peak.

	Will contain NaN where the compound was not in the hit list for a peak.
	"""

	hit_numbers: List[int] = attr.field(default=attr.Factory(list))
	"""
	List of "hit" numbers from NIST MS Search.

	Lower is better. Will contain NaN where the compound was not in the hit list for a peak.
	"""

	#: The reference mass spectrum for the compound from the NIST library.
	reference_data: Optional[ReferenceData] = attr.field(converter=_attrs_convert_reference_data, default=None)

	@property
	def match_factor(self) -> float:
		"""
		The average match factor.

		Missing values (where the compound was not in the hit list for a peak) are excluded from the calculation.
		"""

		return numpy.nanmean(self.mf_list)  # type: ignore[return-value]  # mypy thinks type is "floating[Any]"

	@property
	def match_factor_stdev(self) -> float:
		"""
		The standard deviation of the match factors.

		Missing values (where the compound was not in the hit list for a peak) are excluded from the calculation.
		"""

		return numpy.nanstd(self.mf_list)  # type: ignore[return-value]  # mypy thinks type is "floating[Any]"

	@property
	def reverse_match_factor(self) -> float:
		"""
		The average reverse match factor.

		Missing values (where the compound was not in the hit list for a peak) are excluded from the calculation.
		"""

		return numpy.nanmean(self.rmf_list)  # type: ignore[return-value]  # mypy thinks type is "floating[Any]"

	@property
	def reverse_match_factor_stdev(self) -> float:
		"""
		The standard deviation of the reverse match factors.

		Missing values (where the compound was not in the hit list for a peak) are excluded from the calculation.
		"""

		return numpy.nanstd(self.rmf_list)  # type: ignore[return-value]  # mypy thinks type is "floating[Any]"

	@property
	def average_hit_number(self) -> float:
		"""
		The average hit number.

		Missing values (where the compound was not in the hit list for a peak) are excluded from the calculation.
		"""

		return numpy.nanmean(self.hit_numbers)  # type: ignore[return-value]  # mypy thinks type is "floating[Any]"

	@property
	def hit_number_stdev(self) -> float:
		"""
		The standard deviation of the hit numbers.

		Missing values (where the compound was not in the hit list for a peak) are excluded from the calculation.
		"""

		return numpy.nanstd(self.hit_numbers)  # type: ignore[return-value]  # mypy thinks type is "floating[Any]"

	def __len__(self) -> int:
		"""
		The number of aligned peaks the compound appeared in the hit list for.
		"""

		return numpy.count_nonzero(~numpy.isnan(self.hit_numbers))

	def __repr__(self) -> str:
		return f"<Consolidated Search Result: {self.name} \tmf={self.match_factor}\tn={len(self)}>"

	def __str__(self) -> str:
		return self.__repr__()

	def to_dict(self) -> Dict[str, Any]:
		"""
		Returns a dictionary representation of this :class:`~.ConsolidatedSearchResult`.

		All keys are native, JSON-serializable, Python objects.
		"""

		if self.reference_data is None:
			reference_data_as_dict = None
		else:
			reference_data_as_dict = self.reference_data.to_dict()

		return {
				"name": self.name,
				"cas": self.cas,
				"mf_list": self.mf_list,
				"rmf_list": self.rmf_list,
				"hit_numbers": self.hit_numbers,
				"reference_data": reference_data_as_dict,
				}

	@classmethod
	def from_dict(cls: Type["ConsolidatedSearchResult"], d: Mapping[str, Any]) -> "ConsolidatedSearchResult":
		"""
		Construct a :class:`~.ConsolidatedSearchResult` from a dictionary.

		:param d:
		"""

		hit_numbers = [float("nan") if hn == -65535 else hn for hn in d["hit_numbers"]]
		mf_list = [float("nan") if mf == -65535 else mf for mf in d["mf_list"]]
		rmf_list = [float("nan") if mf == -65535 else mf for mf in d["rmf_list"]]

		return cls(
				name=d["name"],
				cas=d["cas"],
				mf_list=mf_list,
				rmf_list=rmf_list,
				hit_numbers=hit_numbers,
				reference_data=d["reference_data"],
				)


def _attrs_convert_hits(hits: Optional[List[ConsolidatedSearchResult]]) -> List[ConsolidatedSearchResult]:
	if hits is None:
		hits = []

	hits = list(hits)
	for hit in hits:
		if not isinstance(hit, ConsolidatedSearchResult):
			raise TypeError("'hits' must be a list of ConsolidatedSearchResult objects")

	return hits


@attr.define(init=False)
class ConsolidatedPeak:
	"""
	A Peak that has been produced by consolidating the properties and search results of several qualified peaks.

	:param rt_list: List of retention times of the aligned peaks.
	:param area_list: List of peak areas for the aligned peaks.
	:param ms_list: List of mass spectra for the aligned peaks.
	:param minutes: Retention time units flag.
		If :py:obj:`True`, retention time is in minutes;
		if :py:obj:`False` retention time is in seconds
	:param hits: Optional list of possible identities for this peak.
	:param ms_comparison: Mapping or Pandas :class:`~pandas.Series` giving pairwise mass spectral comparison scores.
	:param meta: Optional dictionary for storing e.g. peak number or whether the peak should be hidden.

	.. latex:clearpage::
	.. autosummary-widths:: 3/10
	"""

	#: List of retention times of the aligned peaks.
	rt_list: List[float] = attr.field(converter=list)

	#: List of peak areas for the aligned peaks.
	area_list: List[float] = attr.field(converter=list)

	#: List of mass spectra for the aligned peaks.
	ms_list: MutableSequence[Optional[MassSpectrum]] = attr.field(converter=list)

	#: Optional list of possible identities for this peak.
	hits: List[ConsolidatedSearchResult] = attr.field(converter=_attrs_convert_hits)

	#: Pairwise mass spectral comparison scores.
	ms_comparison: pandas.Series = attr.field(converter=_attrs_convert_ms_comparison)

	#: Optional dictionary for storing e.g. peak number or whether the peak should be hidden.
	meta: Dict[str, Any]

	def __init__(
			self,
			rt_list: List[float],
			area_list: List[float],
			ms_list: MutableSequence[Optional[MassSpectrum]],
			*,
			minutes: bool = False,
			hits: Optional[List[ConsolidatedSearchResult]] = None,
			ms_comparison: Union[Mapping[str, float], pandas.Series, None] = None,
			meta: Optional[Dict[str, Any]] = None
			):

		# TODO: Type check rt_list and ms_list
		# 		if not isinstance(rt, (int, float)):
		# 			raise TypeError("'rt' must be a number")
		#
		# 			if ms and not isinstance(ms, MassSpectrum) and not isinstance(ms, (int, float)):
		# 				raise TypeError("'ms' must be a number or a MassSpectrum object")
		#

		if minutes:
			rt_list = [rt * 60.0 for rt in rt_list]

		self.rt_list = rt_list
		self.area_list = area_list
		self.ms_list = ms_list
		self.meta = meta or {}
		self.ms_comparison = ms_comparison

		if hits is None:
			self.hits = []
		else:
			self.hits = hits

	@property
	def rt(self) -> float:
		"""
		The average retention time across the aligned peaks.
		"""

		return numpy.nanmean(self.rt_list)  # type: ignore[return-value]  # mypy thinks type is "floating[Any]"

	@property
	def rt_stdev(self) -> float:
		"""
		The standard deviation of the retention time across the aligned peaks.
		"""

		return numpy.nanstd(self.rt_list)  # type: ignore[return-value]  # mypy thinks type is "floating[Any]"

	@property
	def area(self) -> float:
		"""
		The average peak area across the aligned peaks.
		"""

		return numpy.nanmean(self.area_list)  # type: ignore[return-value]  # mypy thinks type is "floating[Any]"

	@property
	def area_stdev(self) -> float:
		"""
		The standard deviation of the peak area across the aligned peaks.
		"""

		return numpy.nanstd(self.area_list)  # type: ignore[return-value]  # mypy thinks type is "floating[Any]"

	@property
	def average_ms_comparison(self) -> float:
		"""
		The average of the pairwise mass spectral comparison scores.
		"""

		# return self._average_ms_comparison

		if self.ms_comparison.empty:
			return 0
		else:
			return numpy.nanmean(self.ms_comparison)

	@property
	def ms_comparison_stdev(self) -> float:
		"""
		The standard deviation of the pairwise mass spectral comparison scores.
		"""

		# return self._ms_comparison_stdev

		if self.ms_comparison.empty:
			return 0
		else:
			return numpy.nanstd(self.ms_comparison)

	# def _calculate_spectra(self):
	# 	"""
	# 	Calculate Combined and Averaged spectra
	# 	"""

	# 	mass_lists = []
	# 	intensity_lists = []

	# 	for spec in self._ms_list:

	# 		if spec:
	# 			# print(spec.mass_list)
	# 			mass_lists.append(spec.mass_list)
	# 			intensity_lists.append(spec.intensity_list)
	# 		else:
	# 			# print()
	# 			pass

	# 	if all_equal(mass_lists):
	# 		mass_list = mass_lists[0]
	# 		# print(intensity_lists)
	# 		combined_intensity_list = list(sum(map(numpy.array, intensity_lists)))
	# 		self._combined_mass_spectrum = MassSpectrum(
	# 				mass_list=mass_list, intensity_list=combined_intensity_list
	# 				)

	# 		# averaged_intensity_list = [intensity / len(mass_lists) for intensity in combined_intensity_list]

	# 		averaged_intensity_list = []
	# 		avg_intensity_array = numpy.array(intensity_lists)
	# 		for column in avg_intensity_array.T:
	# 			if sum(column) == 0 or numpy.count_nonzero(column) == 0:
	# 				averaged_intensity_list.append(0)
	# 			else:
	# 				averaged_intensity_list.append(sum(column) / numpy.count_nonzero(column))

	# 		self._averaged_mass_spectrum = MassSpectrum(
	# 				mass_list=mass_list, intensity_list=averaged_intensity_list
	# 				)

	# 	else:
	# 		warnings.warn("Mass Ranges Differ. Unable to process")
	# 		self._combined_mass_spectrum = None
	# 		self._averaged_mass_spectrum = None

	# @property
	# def combined_mass_spectrum(self):
	# 	return self._combined_mass_spectrum

	# @property
	# def averaged_mass_spectrum(self):
	# 	return self._averaged_mass_spectrum

	def __repr__(self) -> str:
		return f"<Consolidated Peak: {self.rt}>"

	def __str__(self) -> str:
		return self.__repr__()

	# def __eq__(self, other):
	# 	"""
	# 	Return whether this ConsolidatedPeak object is equal to another object

	# 	:param other: The other object to test equality with
	# 	:type other: object

	# 	:rtype: bool
	# 	"""

	# 	if isinstance(other, self.__class__):
	# 		if self.rt_list == other.rt_list and self.area_list == other.area_list:
	# 			return self._ms_list == other._ms_list
	# 		#: TODO: compare hits, meta and ms_comparison
	# 		return False
	# 	else:
	# 		return NotImplemented

	def to_dict(self) -> Dict[str, Any]:
		"""
		Returns a dictionary representation of this :class:`~.ConsolidatedPeak`.

		All keys are native, JSON-serializable, Python objects.
		"""

		return {
				"rt_list": self.rt_list,
				"area_list": self.area_list,
				"meta": self.meta,
				"hits": [hit.to_dict() for hit in self.hits],
				"ms_list": [dict(ms) if ms else None for ms in self.ms_list],
				"ms_comparison": self.ms_comparison.to_dict(),
				}

	def __iter__(self) -> Iterator[Tuple[str, Any]]:
		yield from self.to_dict().items()

	def __len__(self) -> int:
		"""
		How many instances of the peak make up this :class:`~.ConsolidatedPeak`.
		"""

		return numpy.count_nonzero(~numpy.isnan(self.rt_list))

	@classmethod
	def from_dict(cls: Type["ConsolidatedPeak"], d: Mapping[str, Any]) -> "ConsolidatedPeak":
		"""
		Construct a :class:`~.ConsolidatedPeak` from a dictionary.

		:param d:
		"""

		hits = []
		for hit in d["hits"]:
			hits.append(ConsolidatedSearchResult.from_dict(hit))

		return cls(
				rt_list=d["rt_list"],
				area_list=d["area_list"],
				ms_list=[MassSpectrum.from_dict(msd) for msd in d["ms_list"]],
				meta=d["meta"],
				ms_comparison=d["ms_comparison"],
				hits=hits,
				)


def pairwise_ms_comparisons(alignment: Alignment, parallel: bool = True) -> pandas.DataFrame:
	"""
	Between Samples Spectra Comparison.

	:param alignment:
	:param parallel: Set to :py:obj:`False` to disable parallelisation.

	:returns: :class:`pandas.DataFrame` where the columns are pairwise spectrum similarity scores and the rows are the peaks.
	"""

	perms: PermsListType = []

	ms_alignment: pandas.DataFrame = alignment.get_ms_alignment(require_all_expr=False)

	# Alternatively ms_alignment.columns
	for i in permutations(alignment.expr_code, 2):
		if i[::-1] not in perms:
			perms.append(i)  # type: ignore[arg-type]

	rows_list: List[Tuple[int, pandas.Series, PermsListType]] = []

	for peak_number, spectra in ms_alignment.iterrows():
		# Spectra is a Series where each element (column) corresponds to a MassSpectrum in a repeat.
		rows_list.append((peak_number, spectra, perms))

	if parallel:
		with Pool(len(ms_alignment.columns)) as p:
			ms_comparison = p.starmap(_map_func, rows_list)
	else:
		ms_comparison = [_map_func(*row) for row in rows_list]

	# TODO: linear mode

	# Convert list of (peak_number, comparisons) pairs to data frame
	column_labels = ["{} & {}".format(*perm) for perm in perms]
	comparison_dict = dict(ms_comparison)

	comparison_df = pandas.DataFrame.from_dict(data=comparison_dict, columns=column_labels, orient="index")

	return comparison_df


def match_counter(
		engine: pyms_nist_search.Engine,
		peak_numbers: List[int],
		qualified_peaks: List[List[QualifiedPeak]],
		ms_comp_data: pandas.DataFrame,
		) -> List[ConsolidatedPeak]:
	"""
	Find the most likely compound for each peak.

	:param engine:
	:param peak_numbers: List of peak numbers to process.
	:param qualified_peaks: List of lists of qualified aligned peaks for each repeat.
	:param ms_comp_data: Dataframe giving pairwise mass spectrum comparisons for each set of aligned peaks.
	"""

	# Convert peak_numbers to a set and sort smallest to largest
	peak_numbers = sorted(set(peak_numbers))

	peak: Optional[QualifiedPeak]
	hit: SearchResult

	aligned_peaks = []
	consolidated_peaks = []

	for idx, peak_no in enumerate(peak_numbers):
		# idx relates to the position in the QualifiedPeak lists for each repeat.
		# peak_no correlates to the ``peak_number`` attribute of the QualifiedPeak objects
		row: List[Optional[QualifiedPeak]] = []
		# for experiment in project.alignment.expr_code:
		for experiment in qualified_peaks:
			assert experiment is not None
			for peak in experiment:
				if peak.peak_number == peak_no:
					row.append(peak)
					break
			else:
				row.append(None)

		aligned_peaks.append(row)

		rt_data = []
		area_data = []
		ms_data: MutableSequence[Optional[MassSpectrum]] = []
		hits = []
		names = []

		for peak in row:
			if peak:
				rt_data.append(peak.rt)
				assert peak.area is not None
				area_data.append(peak.area)
				ms_data.append(peak.mass_spectrum)

				for hit in peak.hits:
					hits.append(hit)
					names.append(hit.name)

			else:
				rt_data.append(numpy.nan)
				area_data.append(numpy.nan)
				ms_data.append(None)

		names.sort()
		names_count = Counter(names)
		# print(names_count)
		# exit()

		hits_data = []

		# for compound, count in names_count.items():
		for compound in names_count:
			# TODO: is the counter actually necessary
			# as the output from the for loop is sorted afterwards anyway?

			# numpy,nan is officially a float, but it doesn't matter for our purposes
			# and the other elements in the lists most definately want to be integers
			NaN = cast(int, numpy.nan)

			mf_data: List[int] = []
			rmf_data: List[int] = []
			hit_num_data: List[int] = []

			for peak in row:
				if peak is None:
					mf_data.append(NaN)
					rmf_data.append(NaN)
					hit_num_data.append(NaN)

				else:
					# print(peak.bounds)
					# input(">")
					for hit_idx, hit in enumerate(peak.hits):
						if hit.name == compound:
							mf_data.append(hit.match_factor)
							rmf_data.append(hit.reverse_match_factor)
							hit_num_data.append(hit_idx + 1)
							CAS = hit.cas
							spec_loc = hit.spec_loc

							break
					else:
						mf_data.append(NaN)
						rmf_data.append(NaN)
						hit_num_data.append(NaN)

			# print(f"Obtaining reference data for {compound} (CAS {CAS})")
			ref_data = engine.get_reference_data(spec_loc)
			# print(ref_data)
			hits_data.append(
					ConsolidatedSearchResult(
							name=compound,
							cas=CAS,
							mf_list=mf_data,
							rmf_list=rmf_data,
							hit_numbers=hit_num_data,
							reference_data=ref_data,
							)
					)

		# Sort consolidated hit list
		hits_data = sorted(hits_data, key=lambda k: (len(k), k.match_factor, -k.average_hit_number), reverse=True)
		# consolidated_peak = ConsolidatedPeak(rt_data, area_data, ms_data, peak_number=n, hits=hits_data, ms_comparison=ms_comp_data.loc[n])
		consolidated_peak = ConsolidatedPeak(
				rt_data,
				area_data,
				ms_data,
				hits=hits_data,
				ms_comparison=ms_comp_data.loc[peak_no],
				meta={"peak_number": idx}  # The position of the peaks in the QualifiedPeak lists
				)

		# consolidated_peak.hits = hits_data  # [:n_hits]

		# consolidated_peak.ms_comparison = ms_comp_data.loc[n]

		consolidated_peaks.append(consolidated_peak)

	return consolidated_peaks


@_fix_init_annotations
@attr.define
class ConsolidatedPeakFilter:
	"""
	Class to filter a list of consolidated peaks to exclude peaks by hit name, match factor etc.

	.. versionadded:: 0.2.0
	"""

	name_filter: List[str] = attr.field(converter=_to_list, default=attr.Factory(list))
	"""
	List of glob-style matches for compound names.

	Consolidated peaks matching any of these will be excluded.
	"""

	min_match_factor: int = Integer.field(default=600)
	"""
	Minimum average match factor.

	Consolidated peaks with an average match factor below this will be excluded.
	"""

	min_appearances: int = Integer.field(default=-1)
	"""
	Number of times the hit must appear across the individual aligned peaks.

	Consolidated peaks where the most common hit appears fewer times than this will be excluded.

	If set to ``-1`` the number of instances of the peak in the project are used.
	"""

	#: If :py:obj:`True` details of excluded peaks will be printed.
	verbose: bool = False

	@classmethod
	def from_method(cls: Type["ConsolidatedPeakFilter"], method: ConsolidateMethod) -> "ConsolidatedPeakFilter":
		"""
		Construct a :class:`~.ConsolidatedPeakFilter` from a :class:`~.ConsolidateMethod`.

		:param method:

		:rtype:

		.. latex:clearpage::
		"""

		return cls(
				name_filter=method.name_filter,
				min_match_factor=method.min_match_factor,
				min_appearances=method.min_appearances,
				)

	def print_skip_reason(self, peak: ConsolidatedPeak, reason: str) -> None:
		"""
		Print the reason for skipping a peak, if :py:attr:`.ConsolidatedPeakFilter.verbose` is :py:obj:`True`.

		:param peak: The peak being skipped.
		:param reason: The reason for skipping the peak.
		"""

		if self.verbose:
			print(f"Skipping peak at {peak.rt / 60:0.3f} mins:", reason)

	def should_filter_peak(self, peak: ConsolidatedPeak) -> bool:
		"""
		Returns :py:obj:`True` if the peak should be excluded based on the current filter options.

		:param peak:
		"""

		hit = peak.hits[0]

		if self.min_appearances == -1:
			# Set to how many times the peak appears
			min_appearances = len(peak)
		else:
			min_appearances = self.min_appearances

		if len(hit) < min_appearances:
			self.print_skip_reason(peak, f"top hit {hit.name!r} only appears {len(hit)} times")
			return True

		hit_name = hit.name.lower()
		for nf in self.name_filter:
			if fnmatch(hit_name, nf):
				self.print_skip_reason(peak, f"name {hit.name!r} matches {nf!r}")
				return True

		# mean_match_factor = mean(hit.mf_list)
		mean_match_factor = hit.match_factor
		if mean_match_factor < self.min_match_factor:
			self.print_skip_reason(peak, f"MF {mean_match_factor} <600")
			return True

		return False

	def filter(self, consolidated_peaks: List[ConsolidatedPeak]) -> List[ConsolidatedPeak]:  # noqa: A003  # pylint: disable=redefined-builtin
		"""
		Filter a list of consolidated peaks.

		:param consolidated_peaks:
		"""

		return [cp for cp in consolidated_peaks if not self.should_filter_peak(cp)]


@_fix_init_annotations
@attr.define
class InvertedFilter(ConsolidatedPeakFilter):
	"""
	Inverted version of :class:`~.ConsolidatedPeakFilter`.

	Returns peaks which would be excluded by a :class:`~.ConsolidatedPeakFilter`.

	.. versionadded:: 0.10.0
	"""

	def print_skip_reason(self, peak: ConsolidatedPeak, reason: str) -> None:  # noqa: D102
		if self.verbose:
			print(f"Would reject peak at {peak.rt / 60:0.3f} mins:", reason)

	def filter(self, consolidated_peaks: List[ConsolidatedPeak]) -> List[ConsolidatedPeak]:  # noqa: A003  # pylint: disable=redefined-builtin
		"""
		Filter a list of consolidated peaks.

		:param consolidated_peaks:
		"""

		filtered_consolidated_peaks = []

		for cp in consolidated_peaks:
			if self.should_filter_peak(cp):
				filtered_consolidated_peaks.append(cp)

		return filtered_consolidated_peaks


def combine_spectra(peak: ConsolidatedPeak) -> Tuple[List[int], List[float]]:
	"""
	Sum the intensities across all mass spectra in the given peak.

	:param peak:

	:returns: List of masses and list of corresponding intensities.

	.. versionadded:: v0.11.0
	"""

	combined_ms_data: Dict[int, float] = defaultdict(float)

	for ms in peak.ms_list:
		if ms is not None:
			for mass, intensity in zip(ms.mass_list, ms.intensity_list):
				combined_ms_data[mass] += intensity

	mass_list, intensity_list = [], []
	for mass, intensity in combined_ms_data.items():
		mass_list.append(mass)
		intensity_list.append(intensity)

	# mass_list, intensity_list = zip(*combined_ms_data.items())

	return mass_list, intensity_list



#!/usr/bin/env python3
#
#  _spectra.py
"""
Internal spectrum comparison utilities.
"""
#
#  Copyright © 2020-2023 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import List, Optional, Sequence, Tuple

# 3rd party
from pyms.Spectrum import MassSpectrum

# this package
from libgunshotmatch.utils import ms_comparison

PermsListType = List[Tuple[int, int]]


def _row_ms_comparison(ms_data: Sequence[MassSpectrum], perms: PermsListType) -> List[Optional[float]]:
	"""
	Performs pairwise mass spectral comparisons of the mass spectra in ``ms_data``.

	:param ms_data:
	:param perms: List of tuples giving the indices of the specta in ``ms_data`` to compare.

	:returns: List of similarity scores for each permutation.
	"""

	similarity_list = []

	for perm in perms:

		top_ms = ms_data[perm[0]]
		bottom_ms = ms_data[perm[1]]

		similarity_list.append(ms_comparison(top_ms, bottom_ms))

	return similarity_list


def _map_func(
		peak_number: int,
		ms_data: Sequence[MassSpectrum],
		perms: PermsListType,
		) -> Tuple[int, List[Optional[float]]]:
	"""
	Multiprocessing helper function.

	Wraps :func:`~.row_ms_comparison` and also returns the peak number.
	"""

	return peak_number, _row_ms_comparison(ms_data, perms)



#!/usr/bin/env python3
#
#  project.py
"""
Represents a collection of repeat analyses.
"""
#
#  Copyright © 2020-2023 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import os
from typing import Any, Dict, List, Mapping, Optional, Type

# 3rd party
import attr
import pandas  # type: ignore[import-untyped]
import pyms_nist_search
from domdf_python_tools.typing import PathLike
from pyms.DPA.Alignment import Alignment
from pyms.Peak.Class import Peak

# this package
from libgunshotmatch import gzip_util
from libgunshotmatch.consolidate import (
		ConsolidatedPeak,
		ConsolidatedPeakFilter,
		match_counter,
		pairwise_ms_comparisons
		)
from libgunshotmatch.datafile import Repeat
from libgunshotmatch.peak import PeakList, QualifiedPeak, peak_from_dict
from libgunshotmatch.utils import create_alignment

__all__ = ("Project", "consolidate")


@attr.define
class Project:
	"""
	A project represents the aligned peaks from multiple datafiles.
	"""

	#: The name of the project.
	name: str

	#: Peak alignment for the repeats in this project.
	alignment: Alignment

	# datafile_data: Dict[str, DatafileDataElement]

	#: Mapping of repeat names to :class:`~.Repeat` objects.
	datafile_data: Dict[str, Repeat]

	#: List of peaks after :meth:`~.consolidate` is performed. :py:obj:`None` initially.
	consolidated_peaks: Optional[List[ConsolidatedPeak]] = attr.field(default=None)

	def to_dict(self) -> Dict[str, Any]:
		"""
		Returns a dictionary representation of this :class:`~.Project`.

		All keys are native, JSON-serializable, Python objects.
		"""

		alignment_as_dict = {
				"peaks": [PeakList(x).to_list() for x in self.alignment.peakpos],
				"expr_code": self.alignment.expr_code,
				"similarity": self.alignment.similarity,
				}

		if self.consolidated_peaks is None:
			consolidated_peaks_as_list = None
		else:
			consolidated_peaks_as_list = [cp.to_dict() for cp in self.consolidated_peaks]

		datafile_data_as_dict = {k: v.to_dict() for k, v in self.datafile_data.items()}

		return {
				"name": self.name,
				"alignment": alignment_as_dict,
				"datafile_data": datafile_data_as_dict,  # "datafile_data": list(self.datafile_data.keys()),
				"consolidated_peaks": consolidated_peaks_as_list,
				}

	def export(self, output_dir: PathLike) -> str:
		"""
		Export as a ``gsmp`` file.

		:returns: The output filename.
		"""

		export_filename = os.path.join(output_dir, f"{self.name}.gsmp")
		gzip_util.write_gzip_json(export_filename, self.to_dict(), indent=None)
		return export_filename

	@classmethod
	def from_file(cls: Type["Project"], filename: PathLike) -> "Project":
		"""
		Parse a ``gsmp`` file.

		:param filename: The input filename.
		"""

		as_dict: Dict[str, Any] = gzip_util.read_gzip_json(filename)  # type: ignore[assignment]
		return cls.from_dict(as_dict)

	@classmethod
	def from_dict(cls: Type["Project"], d: Mapping[str, Any]) -> "Project":
		"""
		Construct a :class:`~.Project` from a dictionary.

		:param d:
		"""

		alignment_as_dict = d["alignment"]
		alignment_peaks: List[List[Peak]] = []
		for row in alignment_as_dict["peaks"]:
			alignment_peaks.append([])
			for peak in row:
				# print(peak)
				peak_obj = peak_from_dict(peak)

				alignment_peaks[-1].append(peak_obj)

		alignment = create_alignment(
				alignment_peaks,
				alignment_as_dict["expr_code"],
				alignment_as_dict["similarity"],
				)

		consolidated_peaks_as_list = d["consolidated_peaks"]
		if consolidated_peaks_as_list is None:
			consolidated_peaks = None
		else:
			consolidated_peaks = [ConsolidatedPeak.from_dict(cp) for cp in consolidated_peaks_as_list]

		datafile_data = {k: Repeat.from_dict(v) for k, v in d["datafile_data"].items()}

		return cls(
				name=d["name"],
				alignment=alignment,
				consolidated_peaks=consolidated_peaks,
				datafile_data=datafile_data,
				)

	def consolidate(
			self,
			engine: pyms_nist_search.Engine,
			peak_filter: Optional[ConsolidatedPeakFilter] = None,
			) -> pandas.DataFrame:
		"""
		Consolidate the compound identification from the experiments into a single dataset.

		:param engine:
		:param peak_filter: Filter for the consolidated peaks.

		:returns: :class:`pandas.DataFrame` giving the results of pairwise mass spectral comparisons
			between the repeats for each aligned peak.
		"""

		consolidated_peaks, ms_comparison_df = consolidate(self, engine)

		if peak_filter is None:
			self.consolidated_peaks = consolidated_peaks
		else:
			self.consolidated_peaks = peak_filter.filter(consolidated_peaks)

		return ms_comparison_df

		# chart_data = make_chart_data(self)


def consolidate(
		project: Project,
		engine: pyms_nist_search.Engine,
		) -> pandas.DataFrame:
	"""
	Consolidate the compound identification from the experiments into a single dataset.

	:param project:
	:param engine:

	:returns: List of consolidated peaks and :class:`pandas.DataFrame`
		giving the results of pairwise mass spectral comparisons between the repeats for each aligned peak.

	.. versionadded:: 0.10.0
	"""

	ms_comparison_df = pairwise_ms_comparisons(project.alignment)

	peak_numbers: List[int] = []
	peak: Optional[QualifiedPeak]

	qualified_peak_array = []

	# for experiment in project.alignment.expr_code:
	for experiment in project.datafile_data:
		qualified_peaks = project.datafile_data[experiment].qualified_peaks
		assert qualified_peaks is not None
		for peak in qualified_peaks:
			assert peak.peak_number is not None
			peak_numbers.append(peak.peak_number)
		qualified_peak_array.append(qualified_peaks)

	# Convert peak_numbers to a set and sort smallest to largest
	peak_numbers = sorted(set(peak_numbers))

	consolidated_peaks = match_counter(
			engine=engine,
			peak_numbers=peak_numbers,
			qualified_peaks=qualified_peak_array,
			ms_comp_data=ms_comparison_df,
			)

	return consolidated_peaks, ms_comparison_df



#!/usr/bin/env python3
#
#  gzip_util.py
"""
Read and write gzipped JSON.
"""
#
#  Copyright © 2023 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import gzip
from typing import Any, Dict, List, Optional, OrderedDict, Tuple, Union

# 3rd party
import sdjson
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.typing import PathLike

__all__ = ("read_gzip_json", "write_gzip_json")

JSONOutput = Union[str, int, float, bool, None, Dict[str, Any], List[Any]]
JSONInput = Union[JSONOutput, Tuple[Any, ...], OrderedDict]


def read_gzip_json(path: PathLike) -> JSONOutput:
	"""
	Load JSON from a gzipped file.

	:param path: The filename to read from.

	:returns: The loaded JSON content.
	"""

	with gzip.open(PathPlus(path), 'r') as f:
		try:
			# 3rd party
			import orjson
			data = f.read().decode().replace("NaN", "-65535")
			return orjson.loads(data)
		except ImportError:
			return sdjson.load(f)


def write_gzip_json(path: PathLike, data: JSONInput, indent: Optional[int] = 2) -> None:
	"""
	Write JSON to a gzip file.

	:param path: The filename to write to.
	:param data: The JSON-serializable data to output.
	:param indent: Number of spaces used to indent JSON.
	"""

	json_data = sdjson.dumps(data, indent=indent)

	with gzip.open(PathPlus(path), 'w') as f:
		f.write(json_data.encode("utf-8"))
